package controller;

import gui.NewJFrame;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.ArrayDeque;
import java.util.Date;
import java.util.Deque;
import java.util.Iterator;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import javax.swing.SwingUtilities;

import sk.stuba.fiit.tp2010.team20.defence.OpponentDetector;
import sk.stuba.fiit.tp2010.team20.defence.Overtake;
import telemetry.DataStorage;
import telemetry.DataUnit;
import controller.model.Event;
import controller.model.TrackData;
import controller.model.Turn;
import controller.model.TurnDetector;
import controller.model.TickInfo;
import controller.model.Turn.Direction;
import champ2011client.Action;
import champ2011client.Controller;
import champ2011client.SensorModel;

/**
 * Kontroler vozidla zatacajuci podla najdlhsieho senzora.
 */
public class PrototypeFirstTime extends Controller
{	
	DataStorage dS;
	Overtake overtakeMod;	
	//instancia triedy s datami o trati
	TrackData trackData=null;
	//instancia recovery modul
	RecoveryModule recMod;
	//instancia drive modulu
	DriveModul driveMod;
	//instancia Event-ov
	Event event;	
	//9 senzor bez odchylky
	double trackEdgeSenzorStraight;
	
	//radenie
	//zoradeny rad otacok za poslenych 30 tikov
	//private Deque<Double> rpm30Tick = new ArrayDeque<Double>(); 
	private double clutchCur=0;
	//boolean RPMGrows=false;
	boolean notOnTrack =false;
	int dontGear =0;// 7 tikov trva kym otacky zacnu po radeni stupat
	/*
	 * Koeficient frikcie medzi pneumatikou a vozovkou; tato hodnota 
	 * sa v prvom kole na zacitaku zistuje pri rychlosti 150 z brzdnej 
	 * drahy a flagy potrebne pri jej urceni
	 */
	//private double friction=16;
	private boolean kolesaStoja = false,brzdi =false,vypocitajFriction=false;
	private double brakeStartDist;
	//rychlost pri akej sa ma merat frikcia
	private double breakSpeed=125;
	
	// rad sluziaci na komunikaciu s vlaknom vytvarajucim model
	BlockingQueue<TickInfo> tickInfoPipe;

	private boolean trackWidthMeasuring = true;
	int trackWidtgMeasuringCount =8;
	// hovori o tom ci sa po starte dal na stred trate
	private boolean straightMid=false;
	//premenna pre dlzku trate
	private double trackLength=0;
	private boolean trackLengthMeasuring = true;
	
	//premenna pre zaznamenavanie radiusov jednotlivych metrov trate
	private int trackRadius = 1;
	
	private Deque<Integer> turnDirection = new ArrayDeque<Integer>(); 
	
	// potrebne k recovery
	private int ticCount=0;
	private boolean afterStart=false;
	
	//line assist
	private LineAssistModul lnaMod;
	boolean reCount=false;

	// premenne potrebne pre zameranie focusu
	private static Boolean requestedChange = false;
	private static int desiredFocus;
	private static double lastChange;
	private static double focusValue = -1;
	public static Object lock = new Object();
	// premenne pre steering
	private double lastSteer=0;
	private boolean isFirstLap=true;
	boolean rovinaNemeratZakruty=false;
	// premenne pre eventy
	private boolean zmenRychlost = false;
	//rozhodovacia premenna pre launch assist
	boolean lunchAssistSteerOn=true;
	
	FileWriter outFile ;
	PrintWriter out ;
	private String trackName;
	final PrototypeFirstTime me=this;
	
	boolean brzdiuz=false;
	private boolean inLap=true;

	public PrototypeFirstTime()
	{
		tickInfoPipe = new LinkedBlockingQueue<TickInfo>(1000);
		
		Thread t = new Thread(new TurnDetector(tickInfoPipe));
		t.setDaemon(true);
		t.setPriority(Thread.MIN_PRIORITY);
		t.start();

		dS = new DataStorage();
		overtakeMod = new Overtake();
		recMod = new RecoveryModule();
		driveMod = new DriveModul();
		lnaMod = new LineAssistModul();
		event = new Event();
		//PrototypeFirstTime me = this;
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				NewJFrame inst = new NewJFrame(me);
				inst.setLocationRelativeTo(null);
				inst.setVisible(true);
			}
		});
	}
	
	public synchronized void stop(){
		brzdiuz=true;
	}
	public synchronized void start(){
		brzdiuz=false;
	}
	
	public Action control(SensorModel sm)
	{
		//urcuje ci sa ma steering pocita za pomoci launch Assistu
		
		// sleduje sa, ci je uz po starte (100 tickov)
		ticCount++;
		if (ticCount == 100)
			afterStart=true;
		Action action = new Action ();

		checkFocus(sm, action);
		
		double radius;
		//zistenie radiusu
		if ((sm.getDistanceFromStartLine() > trackRadius) && (sm.getDistanceFromStartLine() < (trackRadius+3)))
		{
			trackRadius += 3;
			if (sm.getTrackPosition() > 0)
				radius = computeRadius(sm, 13);
			else
				radius = computeRadius(sm, 3);
			
			if (radius < 2000)
				//System.out.println(radius);
				;
		}
		//odstranenie odchylky zo track edge senzoru s indexom 9
		double straightSensors[] = new double[] {sm.getTrackEdgeSensors()[1],sm.getTrackEdgeSensors()[2],sm.getTrackEdgeSensors()[9],sm.getTrackEdgeSensors()[16],sm.getTrackEdgeSensors()[17]};
		trackEdgeSenzorStraight=SimpleNoisyRemover.average(straightSensors);
		//zistenie sirky trate
		
		//trackData boli nacitane zo suboru
		if(trackData.isLoadData() && isFirstLap){
			isFirstLap=false;
			lnaMod.setTrackWidth(trackData.getTrackWidth());
			trackWidthMeasuring=false;
			driveMod.setTrackWidth(trackData.getTrackWidth());
			driveMod.setFriction(trackData.getFriction());
			event.setEvents(trackData.getEventList());
			event.setJumps(trackData.getJumpList());
			if(!reCount) {
				trackData.reCountEnterPos();
				trackData.reCountBreakDist();
				reCount=true;
			}
		}
		if(!trackData.isLoadData() && sm.getDistanceFromStartLine()<2 && sm.getLastLapTime()!=0) {
			trackData.iterate(sm);
			if(!reCount) {
				trackData.reCountEnterPos();
				trackData.reCountBreakDist();
				reCount=true;
			}
		}
		if(isFirstLap)
			isFirstLap=(sm.getLastLapTime()==0.0);
		
		/*if (trackWidthMeasuring )
		{
			trackData.setTrackWidth(sm.getTrackEdgeSensors()[0]+sm.getTrackEdgeSensors()[18]); 
			//trackData.setTrackWidth(trackWidth);
			//System.out.println(trackWidth+"sirka trate");
			//System.out.println(sm.getTrackPosition());
			driveMod.setTrackWidth(trackData.getTrackWidth());
			lnaMod.setTrackWidth(trackData.getTrackWidth());
			trackWidthMeasuring = false;
		}*/
		if (trackWidthMeasuring && trackWidtgMeasuringCount>=0)
		{
			trackData.setTrackWidth(trackData.getTrackWidth()+sm.getTrackEdgeSensors()[0]+sm.getTrackEdgeSensors()[18]); 
			if(trackWidtgMeasuringCount<8)
				trackData.setTrackWidth(trackData.getTrackWidth()/2);
			System.out.println(trackData.getTrackWidth());
			//trackData.setTrackWidth(trackWidth);
			//System.out.println(trackWidth+"sirka trate");
			//System.out.println(sm.getTrackPosition());
			//driveMod.setTrackWidth(trackWidth);
			if(trackWidtgMeasuringCount==0){
				//trackWidth=(Util.round(trackWidth, 0));
				trackData.setTrackWidth(Util.round(trackData.getTrackWidth(), 0)); 
				driveMod.setTrackWidth(trackData.getTrackWidth());
				lnaMod.setTrackWidth(trackData.getTrackWidth());
				trackWidthMeasuring = false;
			}
			
			trackWidtgMeasuringCount--;
		}
		
		//zistenie dlzky trate
		if (trackLengthMeasuring)
		{
			if ((sm.getDistanceFromStartLine() < trackLength) && (trackLength > 0))
			{
				trackLengthMeasuring = false;
				//System.out.println(trackLength+"dlzka trate");
				dS.setLength(trackLength);
			}
			else
				trackLength=sm.getDistanceFromStartLine();		
		}
		
		// ************ EVENTS and JUMPS **************
		if(sm.getZ() > 0.4 && !isFirstLap)
		{
			event.addJump(sm.getDistanceFromStartLine(), 0, sm.getZ());
			
		}
		if ((sm.getTrackPosition()>1 || recMod.isInRecoveryStateLeft()))
		{
			event.addEvent(sm.getDistanceFromStartLine(), 0, trackData.getNextTurn(), sm);
			if(zmenRychlost)
			{
				event.setSpeed(sm.getDistanceFromStartLine(), sm);
				zmenRychlost = false;
			}
		}
		else
		{
			if ((sm.getTrackPosition()<-1 || recMod.isInRecoveryStateRight()))
			{
				event.addEvent(sm.getDistanceFromStartLine(), 0, trackData.getNextTurn(), sm);
				if(zmenRychlost)
				{
					event.setSpeed(sm.getDistanceFromStartLine(), sm);
					zmenRychlost = false;
				}
			}
			else
			{
				zmenRychlost = true;
			}
		}

		// by paro
		if (sm.getDistanceFromStartLine() < 2){
			trackData.setEventList(event.getEvents());
			trackData.setJumpList(event.getJumps());
			//System.out.println("na zaciatku "+event.getEvents().size());
		}
		
		
		/* * * * * R E C O V E R Y * * * Z A C I A T O K * * */

		/* 
		 * v kazdom ticku sa overi, ci sa pilot nenachadza v specifuckych situaciach recovery 
		 * 
		 */
		
		// ak je po starte, kontorluje sa specificka situacia
		if (afterStart)
			recMod.checkRecovery(sm);
		
		// zablokovany pri okraji trate (kolmo) - prechod do noveho recovery stavu
		if ((recMod.isBlocked()) && afterStart && (-1<sm.getTrackPosition() && sm.getTrackPosition()<1) && !brzdiuz){
			if (sm.getTrackPosition()<0)
				return recMod.recoverRight(sm); 
			else if (sm.getTrackPosition()>0)
				return recMod.recoverLeft(sm);
		}
			
		// zablokovany v kopci 			
		if ((recMod.isBlocked()) && afterStart && !brzdiuz)
			return recMod.goStraight(sm);

		// vyletel som na lavu stranu cesty alebo bol v recovrery state left a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if ((sm.getTrackPosition()>1 || recMod.isInRecoveryStateLeft()) && (!recMod.isBlocked())){
			notOnTrack=true;
			return recMod.recoverLeft(sm);
		} //else if (recMod.isBlocked())
		
		// vyletel som na pravu stranu cesty alebo bol v recovrery state right a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if ((sm.getTrackPosition()<-1 || recMod.isInRecoveryStateRight()) && (!recMod.isBlocked())){
			notOnTrack=true;
			return recMod.recoverRight(sm);
		} //else if (recMod.isBlocked())

		// som na trati v protismere - natoceny viac do lava 
		if ((sm.getAngleToTrackAxis()>Math.PI/2.0) && (sm.getAngleToTrackAxis()<Math.PI)){
//			System.out.println("som na trati v protismere - natoceny viac do lava");
			return recMod.recoverRight(sm);
		}
		
		// som na trati v protismere - natoceny viac do prava
		if ((sm.getAngleToTrackAxis()<-Math.PI/2.0) && (sm.getAngleToTrackAxis()>-Math.PI)){
//			System.out.println("som na trati v protismere - natoceny viac do prava");
			return recMod.recoverLeft(sm);
		}
		
		/* * * * * R E C O V E R Y * * * K O N I E C * * */		
		
		
		/*
		 * rychlost - akceleracia naplno do dosiahnutia maximalnej rychlosti, ktoru je mozne ubrzdit na vzdialenost
		 * najdlhsieho senzora,
		 * vzorec			v = odmocnina(2 * g * mi * brzdna draha)
		 */
		double speedAdjustment;
		
		if(!isFirstLap)
			straightMid=true;
		if(!driveMod.isStraightMid())
			speedAdjustment=120;
		else if(isFirstLap)
			speedAdjustment=60;
		/*else if(nextTurn!=null && (nextTurn.getStartDistance()-sm.getDistanceFromStartLine())<50)
			speedAdjustment=0;
		else if(nextTurn!=null && nextTurn.getStartDistance()<sm.getDistanceFromStartLine())
			speedAdjustment=20;*/
		else if(isInTurn(sm, trackData.getNextTurn()) && trackData.getNextTurn().getPeakValue()>0.17)
			speedAdjustment=20;
		else /*if(trackWidth>12)
			speedAdjustment=-10;
		else if(trackWidth>14)
			speedAdjustment=-20;
		else*/
			speedAdjustment=0;
			
			/*if(sm.getSpeed()>100){
				action.accelerate = 0;
				ESP.actionBrakeAktual = 0;
				ESP.actionBrakeTickBack = 0;
				
				action.brake = 0;
				ESP.actionAccelAktual = 0;
				ESP.actionAccelTickBack = 0;
			}else*/ 
			
		
			
			
			if(trackData.isLoadData() || sm.getLastLapTime()!=0) {
				action=lnaMod.getSpeedFromNextTurn(action, sm, trackData, event.getSpeed(trackData.getActEvent()), trackEdgeSenzorStraight);
				//System.out.println("accel: <"+action.accelerate);
			}
			else {
				if ((sm.getSpeed()+ speedAdjustment + event.getSpeed(trackData.getActEvent()))  < Math.sqrt(2 * 9.81 * driveMod.getFriction() * (Math.abs(trackEdgeSenzorStraight-4)))){
					action.accelerate = ESP.filterASR(sm, ESP.actionAccelAktual);
					ESP.actionBrakeAktual = 0;
					ESP.actionBrakeTickBack = 0;
				}else if(sm.getSpeed()<20){
					action.accelerate = 0.8;
					ESP.actionBrakeAktual = 0;
					ESP.actionBrakeTickBack = 0;
				}
				else{
					action.brake = ESP.filterABS(sm, ESP.actionBrakeAktual);
					ESP.actionAccelAktual = 0;
					ESP.actionAccelTickBack = 0;
				}
			}
		// * riadenie *
		// zmena smeru na zaklade najdlhsieho senzora
		//action.steering = steeringAdjustment(computeAvgLongestSensorAngle(sm.getTrackEdgeSensors()));
		int max =0;
		for (int i = 1; i < sm.getTrackEdgeSensors().length; i++)
			if (sm.getTrackEdgeSensors()[max] < sm.getTrackEdgeSensors()[i] && i!=1 && i!=2 && i!=16 && i!=17)
				max = i;
		
		if(isFirstLap){
			action.steering=driveMod.firstLapSteer(trackEdgeSenzorStraight, sm.getTrackEdgeSensors()[max], sm.getAngleToTrackAxis(), sm.getTrackPosition(),max);
			//System.out.println("first lap steer");
		}
		//System.out.println("odstranenie sumu: "+trackEdgeSenzorStraight);
		
		straightMid=driveMod.isStraightMid();
		//rovinaNemeratZakruty=driveMod.isRovinaNemeratZakruty();
		
		if(isFirstLap);
		//	action.steering=firstLapSteer(sm);
			
		
		
		//LINE ASSIST
		//*************************************/
		//aktualizacia TrackData podla aktualnej polohy
		//*************************************/
		trackData.iterate(sm);
		
		//nieje v prvom kole riadim inak 
		if(!isFirstLap){
			action.steering=driveMod.steerNormal(sm.getTrackEdgeSensors(),sm.getDistanceFromStartLine(), sm.getAngleToTrackAxis(), sm.getTrackPosition(), trackData.getNextTurn(),trackData.getLastTurnEnd(), trackData.isTurnFront());
			//System.out.println("normal steer");
		}
		
		//korekcia steeringu podla typu zakruty
		if(sm.getLastLapTime() !=0 || trackData.isLoadData()) {
			action=lnaMod.assist(sm, action, trackData);
			//System.out.println(trackData.getNextTurn().getPeakValue());
		}
		//LINE ASSIST
		//*************************************/
		//KONIEC
		//*************************************/
		
		
		if(!isInAir(sm))
			action = driveMod.getGear(sm, action);
		else
			action.gear=sm.getGear();

		
		try {
			tickInfoPipe.put(new TickInfo(action, sm, !driveMod.isRovinaNemeratZakruty()));
		} catch (InterruptedException e) {
			//System.err.println("Error: chyba pri posielani akcie do modelu trate.");
			e.printStackTrace();
		}
		//urcuje ci sa ma steering pocita za pomoci launch Assistu pokial nedostahne prvu zakrutu
		if(trackData.isLoadData() && trackData.getNextTurn().getStartDistance()-trackData.getTrackWidth()<sm.getDistanceFromStartLine() && trackData.getNextTurn().getEndDistance()>sm.getDistanceFromStartLine() && lunchAssistSteerOn)
			lunchAssistSteerOn=false;
		if(lunchAssistSteerOn && trackData.isLoadData()){
			action.steering=driveMod.launchAssitSteer(sm,0.0);
		}
		else
		{
			if (trackData.getNextTurn()!=null)
			action = overtakeMod.overtaking(sm, action, trackData.getTrackWidth(), trackData.getNextTurn());
			if (action.brake > 0)
			{
				ESP.actionAccelAktual = 0;
				ESP.actionAccelTickBack = 0;
			}
		}
		
		
		if(driveMod.getFriction() == 16){
			action=driveMod.getFriction(sm.getWheelSpinVelocity(), sm.getSpeed(), sm.getDistanceRaced(), action);
			ESP.actionAccelAktual = 1;
			ESP.actionAccelTickBack = 1;
			trackData.setFriction(driveMod.getFriction());
		}
		
		/*if(friction ==16 ){
			friction=getFriction(sm, action);
			ESP.actionAccelAktual = 1;
			ESP.actionAccelTickBack = 1;
		}*/
	/*if(sm.getSpeed()>100)
			System.out.println(sm.getCurrentLapTime());*/
		//System.out.println(Math.toDegrees(sm.getAngleToTrackAxis()));
		//System.out.println(-computeAvgLongestSensorAngle(sm.getTrackEdgeSensors()));
		//System.out.println("rovina: "+rovinaNemeratZakruty+", uhol: "+sm.getAngleToTrackAxis()+", "+"steer: "+action.steering);
		lastSteer=action.steering;
		
		//System.out.println(sm.getZ());
		
		//getSensor9Avg(sm);
		//for(int k=0;k<sm.getOpponentSensors().length;k++)
			//System.out.print(sm.getOpponentSensors()[k]+", ");
			
		//System.out.println();
		//action=OpponentDetector.monitor(action, sm, trackData);
		//action = simpleOvertakeAdjust(sm, action);
		
		//action = overtakeMod.overtaking(sm, action, trackData.getTrackWidth());
	//	System.out.println("pisem next "+trackData.getNextTurn().getStartDistance());
		if(brzdiuz){
			action.accelerate=0;
			action.brake=1;
			ESP.actionAccelAktual = 0;
			ESP.actionAccelTickBack = 0;
		}
		
		double accel;
		if (action.accelerate > 0)
			accel = action.accelerate;
		else
			accel = -action.brake;
		if (dS.getCount()>0)
		{
			if (((int)sm.getDistanceRaced())>(int)dS.getLast().getDistance())
			{
					dS.store(new DataUnit(accel, sm.getSpeed(), sm.getRPM(), action.gear, sm.getTrackPosition(), sm.getCurrentLapTime(), sm.getDistanceRaced()));
					if ((inLap) && (sm.getDistanceFromStartLine() > 0) && (sm.getDistanceFromStartLine() < 5))
					{
						inLap = false;
						dS.addFirstMeters(dS.getCount()-1);
					}
					if ((!inLap) && (sm.getDistanceFromStartLine() > 5))
						inLap = true;
			}
		}
		else
			dS.store(new DataUnit(accel, sm.getSpeed(), sm.getRPM(), action.gear, sm.getTrackPosition(), sm.getCurrentLapTime(), sm.getDistanceRaced()));

		
		return action;
	}
	
	/**
	 * zisti ci je vozidlo vo vzduchu (ak prekroci 0.4) 
	 * @param sm
	 * @return
	 */
	boolean isInAir(SensorModel sm){
		return (sm.getZ()>0.4);
	}
	
	
	/**
	 * Metoda urcena na vyziadanie si zmeny focusu.
	 * @param desiredFocus zelane nastavenie focusu z rozsahu <-90, 90>
	 */
	public static void requestFocusChange(int desiredFocus) {
		synchronized (requestedChange)
		{
			PrototypeFirstTime.desiredFocus = desiredFocus;
			requestedChange = true;
		}
	}
	/**
	 * 
	 * @param oppSenz
	 * @param trackSenz
	 * @param kriterium hranica v metroch ktora urcuje dlazku pod ktorou sa senzor prekrije 
	 * @return prekryte trackedgesenzor
	 */
	private double[] prekriOpponentSenzorySTrackEdge(double[] oppSenz,double[] trackSenz,double kriterium,SensorModel sm){
		
		double[] pom=new double[19];
		
		for (int i = 9; i < 28; i++) {
			pom[i-9]=oppSenz[i];
		}
		
		
		for(int i=0;i<trackSenz.length;i++){
			if(pom[i]<kriterium ){
				System.out.println("nahradzam: "+trackSenz[i]+" tymto: "+pom[i]+" na indexoch "+i+" a "+(i));
				trackSenz[i]=0;
				if(i>9 && i<17 && sm.getTrackPosition()>-0.9){
					trackSenz[i+1]=0;
					trackSenz[i+2]=0;
				}else if(i<9 && i>1 && sm.getTrackPosition()<0.9){
					trackSenz[i-1]=0;
					trackSenz[i-2]=0;
				}
				else if(i==9){
					trackSenz[i+1]=0;
					trackSenz[i-1]=0;
				}
				else if (i>1 && i<17){
					trackSenz[i+1]=0;
					trackSenz[i-1]=0;
					trackSenz[i+2]=0;
					trackSenz[i-2]=0;
				}
				
			}
		}
		
		return trackSenz;
	}
	/**
	 * metoda riadenie mimo prveho kola ridi sa najdlhsim senzorom
	 * @param sm
	 * @param aktTurn
	 * @return
	 */
	private double steerNormal(SensorModel sm, Turn aktTurn,double lastTurnEnd){
		double steer=0;
		double[] prekrytesenzory = prekriOpponentSenzorySTrackEdge(sm.getOpponentSensors(),sm.getTrackEdgeSensors(),25.0,sm);
		
		double uhol = -computeAvgLongestSensorAngle(prekrytesenzory);
		double[]  maxSteers = {0,0.25, 0.38, 0.48,0.68,0.70,0.85,1,1,1 };
		int maxSteerIndex=(int)Math.abs(uhol)/10;
		//smer zakruty
		int dir = ((int) (Math.abs(uhol)/uhol));
		//System.out.println("smer: "+dir+" laststeer: "+lastSteer+" ");
		//if(aktTurn.getStartDistance()<sm.getDistanceFromStartLine() && aktTurn.getEndDistance()>sm.getDistanceFromStartLine()){
			//System.out.println("smer: "+dir+" laststeer: "+lastSteer+" steer: "+steer+" uhol: "+uhol);
			if(maxSteers[maxSteerIndex]>Math.abs(lastSteer) ){
				steer=(Math.abs(lastSteer)+0.01*maxSteerIndex)*dir;
			}else 
				steer=(Math.abs(lastSteer)-0.01*maxSteerIndex)*dir;
		//}else if(aktTurn.getStartDistance()-sm.getDistanceFromStartLine()>200 && aktTurn.getPeakValue()>0.25 || aktTurn.getPeakValue()<0.25 && aktTurn.getStartDistance()-sm.getDistanceFromStartLine()>50)
		//	steer=(sm.getAngleToTrackAxis())/0.366519;
		//}else 
		//	steer=0;
		//System.out.println("smer: "+dir+" laststeer: "+lastSteer+" steer: "+steer+" uhol: "+uhol);
		if(lastSteer==0 && steer!=0 && isInTurn(sm,aktTurn))
			steer=aktTurn.getPeakValue();
			
		if(steer == 0 && lastSteer == 0 && aktTurn.getStartDistance()>sm.getDistanceFromStartLine() && aktTurn.getStartDistance()-sm.getDistanceFromStartLine()>2*trackData.getTrackWidth()){
			if(sm.getTrackPosition()>0.5)
				steer=(sm.getAngleToTrackAxis()-0.05)/0.366519;
			else if(sm.getTrackPosition()<-0.5)
				steer=(sm.getAngleToTrackAxis()+0.05)/0.366519;
			
			
		}
		if(uhol ==0 && lastSteer!=0){
			steer=(Math.abs(lastSteer)-0.03)*-(Math.abs(lastSteer)/lastSteer);
			if(lastSteer<0.03 && lastSteer>-0.03)
				steer=0;
		}
			
		//if(steer>aktTurn.getPeakValue())
		//	steer=aktTurn.getPeakValue();
		
		/*if(aktTurn.getStartDistance()<sm.getDistanceFromStartLine() && aktTurn.getEndDistance()>sm.getDistanceFromStartLine() && aktTurn.getPeakValue()<0.15){
			if(aktTurn.getDirection()==Direction.RIGHT)
				steer=(sm.getAngleToTrackAxis()-(sm.getTrackPosition()+0.95)*0.2)/0.366519;
			if(aktTurn.getDirection()==Direction.LEFT)
				steer=(sm.getAngleToTrackAxis()-(sm.getTrackPosition()-0.95)*0.2)/0.366519;
		}
		*/if(aktTurn.getStartDistance()-lastTurnEnd>30 && aktTurn.getStartDistance()-5>sm.getDistanceFromStartLine() && lastTurnEnd<sm.getDistanceFromStartLine() || lastTurnEnd>aktTurn.getStartDistance()){
			steer=(sm.getAngleToTrackAxis()-(sm.getTrackPosition())*0.1)/0.366519;
		}
		if(sm.getAngleToTrackAxis()>0.8722 && sm.getAngleToTrackAxis()<-0.8722)
			steer=1*dir;
		
		return steer;
		
	}
	/**
	 * zisti ci je v zakrute v parametri
	 * @param sm
	 * @param aktTurn
	 * @return
	 */
	private boolean isInTurn(SensorModel sm, Turn aktTurn){
		if(aktTurn!=null)
			return aktTurn.getStartDistance()<sm.getDistanceFromStartLine() && aktTurn.getEndDistance()>sm.getDistanceFromStartLine();
		return false;
	}
	/**
	 * metoda na nastavanie steeringu v prvom kole, zatial sa vyuziva ja v ostatnych kolach,
	 * okrem to ho ze vracia steering upravuje a hodnoty globalnych premenny ako rovinaNemeratZakruty, turnDirection 	
	 *	
	 * @param sm senzor model
	 * @return vracia steerin 
	 */
	private double firstLapSteer(SensorModel sm){
		//TODO vynulovat que smeru po zmene smeru
		
		double steer=0;
		int max=0;
		for (int i = 1; i < sm.getTrackEdgeSensors().length; i++)
			if (sm.getTrackEdgeSensors()[max] < sm.getTrackEdgeSensors()[i] && i!=1 && i!=2 && i!=16 && i!=17)
				max = i;
		int uhol= 0;	
		//boolean pre vytvaranie modelu 
		if(straightMid && (sm.getAngleToTrackAxis()<-0.001 || sm.getAngleToTrackAxis()>0.001)  && (sm.getTrackPosition()>0.01 || sm.getTrackPosition()<-0.01))
			rovinaNemeratZakruty=false;
		
		//System.out.println("lavo: "+sm.getTrackEdgeSensors()[8]+"stredny: "+sm.getTrackEdgeSensors()[9]+" pravo: "+sm.getTrackEdgeSensors()[10]);
		
		if(max<9  && ((sm.getTrackEdgeSensors()[max]-sm.getTrackEdgeSensors()[9])>trackData.getTrackWidth() || sm.getTrackEdgeSensors()[max]<2*trackData.getTrackWidth() || turnDirection.getFirst()!=0) ){
			uhol=1;
			
		}
		else if(max>9 && ((sm.getTrackEdgeSensors()[max]-sm.getTrackEdgeSensors()[9])>trackData.getTrackWidth() || sm.getTrackEdgeSensors()[max]<2*trackData.getTrackWidth() || turnDirection.getFirst()!=0) ){
			uhol=-1;
			
		}
		else{
			uhol =0;
		}
		//uhol v predchadzjucom tiku prenasobeny -1 aby pri porovnani s uhlom v akt tiku bol rovnaky ak bola preskocena nula
		int uholLast=getTurnDirection(turnDirection)*-1;
		if(turnDirection.size()==5)
			turnDirection.pollFirst();
		turnDirection.addLast(uhol);
		uhol =getTurnDirection(turnDirection);
		//System.out.println(uhol);
		//zenil sa smer zakruty bez toho aby bola zaregistrovana 0
		if(uhol==uholLast){
			rovinaNemeratZakruty=true;
			uhol =0;
		}
		//zaregistroval som smer rovina
		else if(uhol==0){
			rovinaNemeratZakruty=true;
			straightMid=false;
		}else if(uhol!=0)//som v zakrute nastavim straightMid na true aby som zakrutu registroval
			straightMid=true;
		
		//if(!rovinaNemeratZakruty)
			//System.out.println(sm.getAngleToTrackAxis());
		if(!rovinaNemeratZakruty && straightMid){
			//if(!(sm.getAngleToTrackAxis()>-0.001 && sm.getAngleToTrackAxis()<0.001)&& uhol !=0 ){
				double straightSensors[] = new double[] {sm.getTrackEdgeSensors()[1],sm.getTrackEdgeSensors()[2],sm.getTrackEdgeSensors()[9],sm.getTrackEdgeSensors()[16],sm.getTrackEdgeSensors()[17]};
				steer=(steeringAdjustmentNew(SimpleNoisyRemover.average(straightSensors),(trackData.getTrackWidth()),0)*(uhol) -(sm.getTrackPosition()*0.1)/0.366519);
				//System.out.println(sm.getTrackEdgeSensors()[1]+" "+sm.getTrackEdgeSensors()[2]+" "+sm.getTrackEdgeSensors()[9]);
			}else {
					steer=(sm.getAngleToTrackAxis()-sm.getTrackPosition()*0.2)/0.366519;
			}
		
		//odstartoval je rovnobezne s osou a na strede trate
		if (!straightMid && (sm.getAngleToTrackAxis()>-0.001 && sm.getAngleToTrackAxis()<0.001) && sm.getTrackPosition()<0.02 && sm.getTrackPosition()>-0.02)
			straightMid=true;
		else
			steer=(sm.getAngleToTrackAxis()-sm.getTrackPosition()*0.5)/0.366519;
		
	
		return steer;
	}
	/** 
	 * 
	 * Override-nuta metoda na nastavenie uhlov senzorov  senzory 1,2,16,17 boli nastaven na smer senzora 9 cize rovnobezne s osou auta, pre zmeny vid metodu isSensorIndexStraight!!!!! 
	 */
	public float[] initAngles()	{
		float[] angles = new float[19];
		for (int i = 0; i < 19; ++i)
			angles[i]=-90+i*10;
		//vid metodu isSensorIndexStraight
		angles[1]=angles[9];
		angles[2]=angles[9];
		angles[17]=angles[9];
		angles[16]=angles[9];
		
		return angles;
	}
	/**
	 * 
	 * @param sm
	 * @return
	 */
	private double getSensor9Avg(SensorModel sm){
		double pom = 0;
		
		out.println(sm.getTrackEdgeSensors()[1]+" "+sm.getTrackEdgeSensors()[2]+" "+sm.getTrackEdgeSensors()[9]+" "+sm.getTrackEdgeSensors()[16]+" "+sm.getTrackEdgeSensors()[17]);
		return pom;
	}
	/**
	 * Metoda sluzi na ziskanie novej hodnoty focusu po zavolani metody
	 * requestFocusChange. Metoda ma byt zavolana z metody control.
	 * @param sm senzor model
	 * @param action vytvarana akcia
	 */
	private void checkFocus(SensorModel sm, Action action) {
		// prestavenie udaju o poslednej zmene pri prejazde startom
		if (sm.getCurrentLapTime() < lastChange)
			lastChange = sm.getCurrentLapTime() - sm.getLastLapTime();
		// nacitanie novej hodnoty focusu a oznamenie cakajucim jej ziskanie
		if (focusValue == -2)
		{
			focusValue = sm.getFocusSensors()[2];
			synchronized (lock)
			{
				lock.notify();
				TurnDetector.setFocusValue(focusValue);
			}
		}
		//zmena focusu
		synchronized (requestedChange)
		{
			if (requestedChange)
				if (sm.getCurrentLapTime() > lastChange + 1) // focus je mozne menit kazdu 1s
				{
					action.focus = desiredFocus;
					requestedChange = false;
					lastChange = sm.getCurrentLapTime();
					focusValue = -2; // -2 vo focusValue znamena, ze chcem novu hodnotu
				}
				else
				{
					requestedChange = false;
					TurnDetector.setFocusValue(-3); // -3 znamena priskore volanie
					synchronized (lock)
					{
						lock.notify();
					}
				}
		}
	}
	
	/**
	 * metoda na zistenie frikcie z brzdnej drahy pri zaseknutych kolesach(vsetky4),
	 *  pouziva v sebe globalne premenne breakSpeed(150), breakStartDist a tri 
	 *  flagy a to vypocitajFriction, brzdi, kolesaStoja
	 * 
	 * @param sensors aktualne senzory
	 * @param action aktualna akcia
	 * @return 0 ak auto este nezastavilo a frikciu pre povrch ak auto zastalo
	 */
    private double getFriction(SensorModel sensors, Action action){
		
		
		if(sensors.getSpeed()<=breakSpeed && !brzdi){
			action.accelerate=1;
			action.brake=0;
			brakeStartDist = sensors.getDistanceRaced();
			//System.out.println(brakeS);
		}else{
			action.steering=0;
			double speed =0;
			//System.out.print(zrychlenie(sensorsBefore.getSpeed(), sensors.getSpeed(), sensorsBefore.getCurrentLapTime(), sensors.getCurrentLapTime()));
			//System.out.print(" =||= ");
			for(int i=0;i<4;i++){
				speed+=sensors.getWheelSpinVelocity()[i]*0.3179;
				//System.out.print(" || "+zrychlenie(sensorsBefore.getWheelSpinVelocity()[i]*0.3179, sensors.getWheelSpinVelocity()[i]*0.3179, sensorsBefore.getCurrentLapTime(), sensors.getCurrentLapTime()));
			}
			speed/=4;
			//System.out.println();
			//System.out.println("rychlost kolies: "+speed +" brakespeed "+breakSpeed+" brakedist "+brakeStartDist);
			
			if(speed<0.05 && !kolesaStoja){
				brakeStartDist= sensors.getDistanceRaced();
				kolesaStoja=true;
				breakSpeed=sensors.getSpeed();
			}
			brzdi=true;
			action.accelerate=0;
			action.brake=1;
			action.clutch=0;
			if(sensors.getSpeed()<0.01 && !vypocitajFriction){
				//System.out.println("drzdna draha"+(sensors.getDistanceRaced()-brakeStartDist)+ " rychlost kolies: "+speed +" brakespeed "+breakSpeed+" brakedist "+brakeStartDist+" friction "+(Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist))));
				vypocitajFriction=true;
				System.out.println(Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist)));
				
				return (Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist)));
				
			}//System.out.println(sensors.getSpeed());
			
		}
		return 16;
	}

	/**
	 * metoda urcujuca polomer useku trate na zaklade troch trackEdge senyorov
	 * 
	 * @param sensors senzormodel
	 * @param first prvy z trojice senzorov, od ktoreho sa pocita radius
	 * @return radius useku trate
	 */
	private double computeRadius(SensorModel sensors, int first)
	{
		double x1 = Math.cos(Math.PI/6) * sensors.getTrackEdgeSensors()[first];
		double y1 = Math.sin(Math.PI/6) * sensors.getTrackEdgeSensors()[first];
		double x2 = Math.cos(Math.PI/4.5) * sensors.getTrackEdgeSensors()[first+1];
		double y2 = Math.sin(Math.PI/4.5) * sensors.getTrackEdgeSensors()[first+1];
		double x3 = Math.cos(Math.PI/3.6) * sensors.getTrackEdgeSensors()[first+2];
		double y3 = Math.sin(Math.PI/3.6) * sensors.getTrackEdgeSensors()[first+2];
		
		double a,b,c,s;
		
		a = Math.sqrt(Math.pow((x1-x2), 2)+Math.pow((y1-y2), 2));
		b = Math.sqrt(Math.pow((x2-x3), 2)+Math.pow((y2-y3), 2));
		c = Math.sqrt(Math.pow((x3-x1), 2)+Math.pow((y3-y1), 2));
		s = (a+b+c)/2;
		
		return (a*b*c)/(4*Math.sqrt(s*(s-a)*(s-b)*(s-c)));
	}
	
	/**
	 * metoda modulu riadenia na urcenie prevodoveho stupna s otacok motora a podla toho ci rastu/klesaju
	 * 
	 * @param sensors senzormodel
	 * @param action akcia na vykonanie
	 * @param rpmGrows boolena hovorici o tom ci otacky klesaju alebo stupaju
	 * @return akciu, ktora ma nastaveny aj prevodovy stupen 
	 */
	private Action getGear(SensorModel sensors,Action action, Deque<Double> rad){
		int gearCur;
		double rpm = sensors.getRPM(),lastRpm=9000;
		int[]  gearUp = {9500, 9500, 9500, 9500, 9500, 0};
		int[]  gearDown =  {0, 4300, 6200, 7000, 7300, 7700};
		gearCur=sensors.getGear();//berem akt stupen
		
		//updatne sa zoznam otacok za poslednych 30 tikov
		if(rad.size()>0)
			lastRpm=rad.getLast();
		if(rad.size()==30 && dontGear==0)
			rad.pollFirst();
		if(dontGear==0)
			rad.addLast(rpm);
		else
			dontGear--;
		
		//som na starte a pretek este nebol odstarovany radim 1, stlacim spojku, pridam plyn 
		if(gearCur<1){
			gearCur=1;
			action.gear=gearCur;
			action.accelerate= 1;
			clutchCur=0.9;
			//dontGear=8;
		}
		//uz je odstartovane a mam stlacenu spojku, zaradenu 1, ak otack rastu a su nad  9000 tak uberam spojku po 0.05, ak klesnu pod 7000 uberam iba po 0.02 ak klesnu pod 6500 pridavam k spojke po 0.01
		//podmienka clutchCur>0.67 a akcia clutchCur-=0.34;
		//TODO ked bude asr s nastavitelni smykom tak zmeni podmienku s casom a dat namiesto acc 1 asr
		else if(((clutchCur>0.05 && gearCur==1 )|| sensors.getSpeed()<55) && sensors.getCurrentLapTime()<3.5 && sensors.getLastLapTime()==0){
			action.accelerate=1;
			//if((rpm-lastRpm)>=0 && rpm > 9300 && sensors.getSpeed()<40)
			//	clutchCur-=0.08;
			//else 
			if((rpm-lastRpm)>=0 && rpm > 9000)
				clutchCur-=0.05;
			else if((rpm-lastRpm)>=0 && rpm > 7000)
				clutchCur-=0.02;
			else if(rpm < 6500)
				clutchCur+=0.01;
			//System.out.println("som tu");
		}else if(clutchCur>0 && (rpm-lastRpm)>0 && rpm>7000 && gearCur==1 && rpm<9499){
			clutchCur-=0.03;
			//System.out.println("lol");
		}
		// rychlost je mensia ako 6 sucastne su otaky vecsie/rovne otackam ked radim, a sucasne otacky v predchadzajucich 30 tikoch stupali => radim hore
		else if (gearCur < 6 && rpm >= gearUp[gearCur-1] && dontGear==0)
		{	
			
			if(gearCur==1)
				clutchCur=0.63;
			if(isDequeGrowing(rad)){
				dontGear=8;
				gearCur++;
			}
			
			
		}
		// rychlost je vetsia ako 1 sucastne su otaky mensie/rovne otackam ked podradzujem a otacky za poslednych 30 tikov klesaju
		else if (gearCur > 1 && rpm <= gearDown[gearCur-1]  && dontGear==0 )
		{
			if(!isDequeGrowing(rad)){
				dontGear=8;
				gearCur--;	
			}
			
		}
			
		//mam zaradenu inu rychlost ako 1 a spojka je stlacena, ak je zaradena dvojka tak pustam spojku ak su otacky nad 6500 ak je ina rychlost tak pustam spojku
		if((clutchCur>0 && gearCur>1 && (rpm-lastRpm)>0) || rpm>9000 && gearCur>1){
			if(gearCur==2 && rpm>7000)
				clutchCur-=0.01;
			else if(gearCur==2 && rpm>8500)
				clutchCur-=0.08;
			else if (gearCur!=2){
				clutchCur-=0.15;
				
			}
		}
		//System.out.println(rpm);
		if(sensors.getCurrentLapTime()<6 && sensors.getLastLapTime()==0)
			action.accelerate=1;
		action.gear=gearCur;
		action.clutch=clutchCur;
		return action;
	}
	/**
	 * metoda vrati true ak rad<double> apon s polovice rastie a false ak nsatane opacne situacia
	 * 
	 * @param que rad cisel ktorych rast sa ma urcit
	 * @return true alebo false
	 */
	private boolean isDequeGrowing(Deque<Double> que){
		
		/*if(que.getLast()-que.getFirst()>0)
			return true;
		else
			return false;*/
		
		Deque<Double> pom = new ArrayDeque<Double>(que);
		boolean rastie[] = new boolean[pom.size()];
		double curValue = pom.pollFirst().doubleValue();
		
		for(int i=0;i<que.size()-1;i++){
			rastie [i]= curValue<pom.peekFirst().doubleValue();
			curValue = pom.pollFirst().doubleValue();
			
		}
		
		int h=0;
		for(int i=0;i<rastie.length;i++){
			if(rastie[i])
				h++;
		}
		
		if(h>=rastie.length/2)
			return true;
		else
			return false;
	}
	/**
	 *  spocita jednotlive prvky na vypocet smeru za poslednych que.size() tikov
	 * @param que rad integerov
	 * @return
	 */
	private int getTurnDirection(Deque<Integer> que){
		Iterator<Integer> i = que.iterator();
		int aktValue=0;
		int zeroCount=0,negCount=0,pozCount=0;
		int ret=0;
		while(i.hasNext()){
			aktValue= i.next();
			ret+=aktValue;
			if(aktValue==0)
				zeroCount++;
			
		}
		if(ret>0)
			return 1;
		else if(ret <0)
			return -1;
		else if(zeroCount>=4)
			return 0;
		else
			return aktValue;
	}	
		
	/**
	 * Metoda urcujuca vztah medzi najdlhsim uhlom a natocenim volantu.
	 * @param longestAngle uhol k najvzdialenejsiemu okraju trate
	 * @return natocenie volantu
	 */
	private double steeringAdjustment(double longestAngle)
	{
		if (Math.abs(longestAngle) > 21)
			if (longestAngle > 0)
				return -1;
			else
				return 1;
		
		return -longestAngle / 21;
	}

	

	/**
	 * Metoda vrati priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 * v rozsahu <-90; 90>. Priemerovana je len jedna suvisla postupnost senzorov.
	 * @param trackEdgeSensors senzory okrajov trate
	 * @return priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 */
	private double computeAvgLongestSensorAngle(double[] trackEdgeSensors)
	{
		int max = 0, numberOfLongestSensors = 1;
		double angle = 0;

		// najde index prveho senzora v smere k najvzdialenejsiemu okraju trate
		for (int i = 1; i < trackEdgeSensors.length; i++)
			
			if (!isSensorIndexStraight(i) && trackEdgeSensors[max] < trackEdgeSensors[i])
					max = i;
		
		// zisti kolko senzorov ma najvyssiu hodnotu
		while (max + numberOfLongestSensors < trackEdgeSensors.length &&
				trackEdgeSensors[max] == trackEdgeSensors[max + numberOfLongestSensors])
			numberOfLongestSensors++;

		// vypocet priemeru, vysledok je z rozsahu <-90; 90>
		for (int i = 0; i < numberOfLongestSensors; i++)
			angle = max + i;
		angle = (angle / numberOfLongestSensors) * 10 - 90;
		
		return angle;
	}
	/**
	 * urci ci index patri medzi indexy ktore su nastavene rovnobezne s osou vozidla metod bola pisna ked tieto indexy boli 1,2,16,17(ak boli zmenene treba zmenit metodu) aj ked je 9 rovnobezne s osou neurcje juako true
	 * @param i
	 * @return
	 */
	private boolean isSensorIndexStraight(int i){
		if(i==1 ||i ==2 || i== 16 || i==17)
			return true;
		else 
			return false;
	}

	/**
	 * Metoda vracajuca vhodnu hodnotu prevodoveho stupna.
	 * 
	 * Implementacia nie je urcena na zaradenie do "produkcneho" kodu.
	 * Jej cielom je uplna oddelenost od ostatnych prvkov ovladaca pre
	 * prehladnost vyvoja.
	 */
	private int setGear(SensorModel sensors)
	{
		int gear = sensors.getGear();

		int[]  gearUp = { 9500, 9400, 9500, 9500, 9500, 0 };
		int[]  gearDown = { 0, 3300, 6200, 7000, 7300, 7700 };
		
		double rpm = sensors.getRPM();
		if (gear < 1)
			gear = 1;

		// check if the RPM value of car is greater than the one suggested
		// to shift up the gear from the current one
		else if (gear < 6 && rpm >= gearUp[gear-1])
			gear++;
		
		// check if the RPM value of car is lower than the one suggested
		// to shift down the gear from the current one
		else if (gear > 1 && rpm <= gearDown[gear-1])
			gear--;
		
		return gear;
	}
	private double radiusFromDistance(double distance){
		if(distance>142)
			return 200;
		else if(distance>44)
			return 0.8107+1.0316*distance-0.0002*Math.pow(distance, 2);
		else if(distance>15)
			return 0.0579+1.0117*distance;
		else
			return 10.0168-0.6658*distance+0.0696*Math.pow(distance, 2);
	}
	private double steeringAdjustmentNew(double distance,double trackWidth,double strana){
		//TODO prve dva ifi na zaklade sirky trate mozno lepsie na zaklde pozicie na trati
		
		distance=distance-(trackWidth-6);
		
		if(distance<42)
			distance= distance*1.05;
		if(distance<15)
			distance= distance*1.1052;
		//System.out.println(distance);
		
		if(distance>138)
			return 0.0075;
		if(distance>43){
			strana= (0.8409/(distance-26.9920))*(1+(strana));
			//strana =0;
		}else if(distance>15){
			strana= (2.2216/(distance-9.9210))*(0.975+(strana));
		}else if (distance >6.3545)
			strana= (3.8193/(distance-6.3542))*(1+(strana));
		else
			strana =1; 
		if(strana>1)
			strana=1;
		if(strana<-1)
			strana=-1;
		
			
		return Math.abs(strana);
	}
	/*vypocita rychlost z steeringu*/
	private double countSpeed(double steering){
		if(steering>0.4){
			return 116.28-114.9683*steering+51.2030*Math.pow(steering, 2);
		}else if(steering>0.14){
			return 189.6921-506.5*steering+582.0934*Math.pow(steering, 2);
		}else if(steering>=0.1){
			return 268.5877-1508.3*steering+3798.4*Math.pow(steering, 2);
		}else if(steering>=0.018){
			return 22.4685/(steering+0.0471);
		}else
			return 500;
			
	}
	public void reset()
	{
		System.out.println("restarting the race");
	}

	public void shutdown()
	{
		//out.close();
		//trackData=new TrackData();
		trackData.setFriction(driveMod.getFriction());
		trackData.setJumpList(event.getJumps());
		trackData.setEventList(event.getEvents());

		//trackData.setEventList(event.events);
		//trackData.setTrackWidth(trackData.getTrackWidth());
		//trackData.setTurnList(turnList);
		try {
			saveTrackData("trackData_"+trackName+".r2w",trackData);
			System.out.println("track data successfully saved into file: "+"trackData_"+trackName+".r2w");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		String directoryName = "telemetry";
		createTelemetryDirIfNeeded(directoryName);
		try {
			directoryName = directoryName.concat(System.getProperty("file.separator")).concat(trackName);
			createTelemetryDirIfNeeded(directoryName);
			String identifier=DateFormat.getDateTimeInstance().format(new Date());
			
			saveTelemetryData(directoryName,identifier.replaceAll(":", "_").concat(".tad"),dS);
			System.out.println("telemetry data successfully saved into file: "+identifier+".tad");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		//dS.visualize();
		System.out.println("shutdown");		
	}
	private void saveTrackData(String path,Object obj) throws FileNotFoundException, IOException{
		File file = new File(path);
		if(!file.exists())
			file.createNewFile();
		ObjectOutput outPut = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream(file)));
		outPut.writeObject(obj);
		outPut.close();
	}
	
	private TrackData readTrackData(String path) throws FileNotFoundException, IOException, ClassNotFoundException{
		File file = new File(path);
		ObjectInput input;
		if(file.exists()){
			input = new ObjectInputStream( new BufferedInputStream( new FileInputStream( file )) );
		
			return (TrackData) input.readObject();
		}
		
		return null;

	}
	public void setTrackName(String trackName) {
		this.trackName = trackName;
		try {
			trackData=readTrackData("trackData_"+trackName+".r2w");
			if(trackData==null) {
				trackData=new TrackData(); 
				trackData.setLoadData(false);
			}
			else {
				trackData.setLoadData(true);
				trackData.reCountEnterPos();
				trackData.reCountBreakDist();
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	private void createTelemetryDirIfNeeded(String directoryName)
	{
		File theDir = new File(directoryName);

		// if the directory does not exist, create it
		if (!theDir.exists())
			theDir.mkdir();
	}
	private void saveTelemetryData(String directoryName, String fileName,Object obj) throws FileNotFoundException, IOException{
		File file = new File(directoryName+System.getProperty("file.separator")+fileName);
		if(!file.exists())
			file.createNewFile();
		ObjectOutput outPut = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream(file)));
		outPut.writeObject(obj);
		outPut.close();
	}
	
}
