package controller;

import java.util.ArrayDeque;
import java.util.Deque;

import champ2011client.Action;
import champ2011client.SensorModel;

/**
 * Trieda zabezpecujuca obnovenie (recovery) tak z miesta mimo trate ako aj na trati
 * v specifickych situacich.
 * 
 * @author Marek Bris
 */

public final class RecoveryModule {
	
	private boolean inRecoveryStateLeft; //=false;
	private boolean inRecoveryStateRight; //=false;
	private boolean blocked;
	private int keepBlock;//=0;
	
	// priemerna rychlost za poslednych 2o tikov
	private Deque<Double> speed20Tick = new ArrayDeque<Double>();
	private double sum=0;
	
	public RecoveryModule(){
		setInRecoveryStateLeft(false);
		setInRecoveryStateRight(false);
		setBlocked(false);
		keepBlock=0;
	}
	
	/**
	 * Kontroluje polohu pilota, ci ho nie je potrebne obnovit, v specifickych situaciach
	 * tak na trati ako mimo nej
	 * @param SensorModel - hodnoty senzorov v aktualnej chvili 
	 * @return Action - nasledujuca akcia
	 */
	
	public void checkRecovery(SensorModel sm){

		// updatne sa priemerna rychlost za poslednych 20 tikov
		if(speed20Tick.size()<20){
			speed20Tick.addLast(Math.abs(sm.getSpeed()));
			sum+=speed20Tick.getLast();
		} else {
			sum-=speed20Tick.getFirst();
			speed20Tick.pollFirst();
			speed20Tick.addLast(Math.abs(sm.getSpeed()));
			sum+=speed20Tick.getLast();
		}
//		System.out.println("speeed > "+sum/20+" keepblock "+keepBlock);
		
	// zabezpecenie "lock-u" recovery mimo trate na rozbehnutie sa 
		
		if ((sum/20 < 0.5) && keepBlock == 0){
			setBlocked(true);
			keepBlock++;
		} else if (keepBlock == 100){
			setBlocked(false);
			keepBlock=0;
		} else if (keepBlock > 0)
			keepBlock++;
	}
	
	
	/**
	 * Prebera kontrolu nad pilotom a dava mu smer jazdy dopredu smerom k trati a to v tom pripade,
	 * ak je v recovery mode a rychlost je nulova
	 * @param SensorModel - hodnoty senzorov v aktualnej chvili 
	 * @return Action - nasledujuca akcia
	 */
	
	public Action goStraight(SensorModel sm){
		Action a=new Action();
		if (sm.getTrackPosition()>1)
			a.steering=-0.8;
		if (sm.getTrackPosition()<-1)
			a.steering=0.8;

		a.gear=1;
		a.accelerate=1;
		a.brake=0;
		return a;
	}
	
	
	/**
	 * Zabezpecuje obnovenia auta z priestoru na lavo od trate, cize navrat na trat
	 * @param SensorModel - hodnoty senzorov v aktualnej chvili 
	 * @return Action - nasledujuca akcia
	 */
	
	public Action recoverLeft(SensorModel sm)
	{
		// P > 1 
		double B=sm.getAngleToTrackAxis();
		Action a= new Action();

		if (sm.getTrackPosition()<-1){
			setInRecoveryStateLeft(false);
			return a;	
		} 
		
		if (isInRecoveryStateRight() && sm.getTrackPosition()>1){
			setInRecoveryStateRight(false);
		} 
		
		a.accelerate = ESP.filterASR(sm, 0.8);
		a.gear=1;
		// som VLAVO
		// rovnobezne s tratou, v smere trate
//			System.out.println("----------- VLAVO "+sm.getSpeed()+" ");
			//System.out.println("rychlost "+sm.getSpeed());
		
			if (-0.785<B && B<0.785){
//			System.out.println("rovnobezne s tratou, v smere trate");
			
			if(sm.getSpeed()<0){
				a.brake=1;
				a.steering=1;
				return a;
			}else {
			double d=(0.436-B)/0.785;
			
			if (d<-1)
				a.steering=-1;
			else
				a.steering=-d;
			// pozicia, kedy to davame controlerovi
			if (B>0)
				setInRecoveryStateLeft(false);
			else {
				setInRecoveryStateLeft(true);
				}
			}
		} 
		// smeruje von z trate
		else if (-2.355<B && B<-0.785){
//		System.out.println("smeruje von z trate");
			a.gear=-1;
			a.steering=1;
			setInRecoveryStateLeft(true);
		}
		// smeruje k stredu trate
		else if (0.785<B && B<2.355){
//		System.out.println("smeruje k stredu trate");
			a.steering=1;
			setInRecoveryStateLeft(true);
		}
		// rovnobezne s tratou, v protismere
		else if ((-2.355>B && B>-3.14) || (2.355<B && B<3.14)){
//		System.out.println("rovnobezne s tratou, v protismere");
			a.steering=1;
			setInRecoveryStateLeft(true);
		}
		return a;
	}
	
	/**
	 * Zabezpecuje obnovenia auta z priestoru na pravo od trate, cize navrat na trat
	 * @param SensorModel - hodnoty senzorov v aktualnej chvili 
	 * @return Action - nasledujuca akcia
	 */
	
	public Action recoverRight(SensorModel sm)
	{
		// P < -1 
		double B=sm.getAngleToTrackAxis();
		Action a= new Action();

		if (sm.getTrackPosition()>1){
			setInRecoveryStateRight(false);
			return a;
		} 
		
		if (isInRecoveryStateLeft() && sm.getTrackPosition()<-1){
			setInRecoveryStateLeft(false);
		} 
		
		a.gear=1;
		a.accelerate = ESP.filterASR(sm, 0.8);

		// som VPRAVO
		// rovnobezne s tratou, v smere trate
		
//		System.out.println("----------- VPRAVO "+sm.getSpeed()+" ");
		if (-0.785<B && B<0.785){
//		System.out.println("rovnobezne s tratou, v smere trate");
		if(sm.getSpeed()<0){
			a.brake=1;
			a.steering=-1;
			return a;
		}else {
			double d=(-0.436-B)/0.785;
			if (d<0) d=-d;
			
			if (d>1)
				a.steering=1;
			else
				a.steering=d;
			// pozicia, kedy to davame controlerovi
			if (B<0)
				setInRecoveryStateRight(false);
			else{
				setInRecoveryStateRight(true);
			}
			}
		} 
		// smeruje von z trate
		else if (0.785<B && B<2.355){
//		System.out.println("smeruje von z trate");
			a.gear=-1;
			a.steering=-1;
			setInRecoveryStateRight(true);
		}
		// smeruje k stredu trate
		else if (-2.355<B && B<-0.785){
//		System.out.println("smeruje k stredu trate");
			a.steering=-1;
			setInRecoveryStateRight(true);
		}
		// rovnobezne s tratou, v protismere
		else if ((-2.355>B && B>-3.14) || (2.355<B && B<3.14)){
//		System.out.println("rovnobezne s tratou, v protismere");
			a.steering=-1;
			setInRecoveryStateRight(true);
		}
	
		return a;
	}

	public void setInRecoveryStateRight(boolean inRecoveryStateRight) {
		this.inRecoveryStateRight = inRecoveryStateRight;
	}

	public boolean isInRecoveryStateRight() {
		return inRecoveryStateRight;
	}

	public void setInRecoveryStateLeft(boolean inRecoveryStateLeft) {
		this.inRecoveryStateLeft = inRecoveryStateLeft;
	}

	public boolean isInRecoveryStateLeft() {
		return inRecoveryStateLeft;
	}

	/**
	 * @param blocked the blocked to set
	 */
	public void setBlocked(boolean blocked) {
		this.blocked = blocked;
	}

	/**
	 * @return the blocked
	 */
	public boolean isBlocked() {
		return blocked;
	}
	
}
