package controller.model;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import controller.DriveModul;
import champ2011client.SensorModel;

/**
 * TrackData je trieda obsahujuca vsetky udaje o trati, vzhladom k tomu ze implementuje serializable je mozne objekty tohto typu ukladat
 * 
 * @author doom
 *
 */
public class TrackData  implements Serializable{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private double friction; //TODO treba dorobit nasetovanie friction po odmerani
	private double trackWidth=0; //sirka trate
	private LinkedList<Turn> turnList=null; //zoznam zakrut
	private LinkedList<EventInfo> eventList;
	private LinkedList<EventInfo> jumpList;
	// TODO
	private EventInfo actEvent=null;
	private int indexEvent=-1;
	private int indexJump=-1;
	private int deviation = 70;	
	boolean isLoadData=false; //ci boli nacitane data zo suboru
	private Turn nextTurn = null; //najblizsia zakruta
	private Turn next2Turn = null; //zakruta ktora nasleduje za najblizsou zakrutou, ak sme pred poslednou zakrutou, hodnota je NULL
	private Turn lastTurn = null; //posledna zakruta
	private int aktTurnIndex=-1; //index aktualnej zakruty v zozname
	private double lastTurnEnd; //koniec poslednej zakruty
	private boolean isTurnFront=false; //ci je nejaka zakruta pred nami - ci sa nachadzame za poslednou zakrutou
	
	public boolean isLoadData() {
		return isLoadData;
	}
	public void setLoadData(boolean trackData) {
		this.isLoadData = trackData;
	}
	public Turn getNextTurn() {
		return nextTurn;
	}
	public void setNextTurn(Turn nextTurn) {
		this.nextTurn = nextTurn;
	}
	public Turn getNext2Turn() {
		return next2Turn;
	}
	public void setNext2Turn(Turn next2Turn) {
		this.next2Turn = next2Turn;
	}
	public Turn getLastTurn() {
		return lastTurn;
	}
	public void setLastTurn(Turn lastTurn) {
		this.lastTurn = lastTurn;
	}
	public double getLastTurnEnd() {
		return lastTurnEnd;
	}
	public void setLastTurnEnd(double lastTurnEnd) {
		this.lastTurnEnd = lastTurnEnd;
	}
	public boolean isTurnFront() {
		return isTurnFront;
	}
	public void setTurnFront(boolean isTurnFront) {
		this.isTurnFront = isTurnFront;
	}
	public double getFriction() {
		return friction;
	}
	public void setFriction(double friction) {
		this.friction = friction;
	}
	public double getTrackWidth() {
		return trackWidth;
	}
	public void setTrackWidth(double trackWidth) {
		this.trackWidth = trackWidth;
	}
	public LinkedList<Turn> getTurnList() {
		return turnList;
	}
	
	public LinkedList<EventInfo> getEventList() {
		return eventList;
	}
public void setEventList(LinkedList<EventInfo> eventList) {
		this.eventList = (LinkedList<EventInfo>) eventList.clone();
		//System.out.println(eventList.size()+"< event");
		this.indexEvent = -1;
	}
	
	public LinkedList<EventInfo> getJumpList() {
		return jumpList;
	}
	public void setJumpList(LinkedList<EventInfo> jumpList) {
		//System.out.println(jumpList.size()+"< jump");
		this.jumpList = (LinkedList<EventInfo>) jumpList.clone();
		this.indexJump = -1;
	}
	
	public boolean isInitTurnList() {
		if(turnList!=null) 
			return true;
		else 
			return false;
	}
	
	/**
	 * prepocitanie vstupnych pozicii
	 */
	public void reCountEnterPos() {
		for (Turn turn : turnList) {
			turn.setEnterPosition(this.trackWidth, this.friction);
		}
	}
	
	/**
	 * prepocitanie breakDist brzdnych vzdialenosti
	 */
	public void reCountBreakDist() { 
		for (Turn turn : turnList) {
			if(turn.getPeakValue()>(trackWidth>14 ? 0.18:0.1))
				turn.setBreakDist(DriveModul.countDistanceFromSpeed(turn.getSpeed(), this.friction));
			else
				turn.setBreakDist(200);
		}
	}
	
	/**
	 * inicializacia zoznamu zakrut z triedy TurnDetector
	 */
	private void initTurnList() {
		this.turnList = TurnDetector.getTurnList(this.trackWidth, this.friction);
		if(turnList.get(0).getEndDistance()>turnList.get(1).getStartDistance()) //odstranenie prvej zakruty ak presahuje ciel
			turnList.remove();
		
		aktTurnIndex=0;
		if(aktTurnIndex<turnList.size()) {						//inicializacia prvej zakruty
			lastTurnEnd=0;
			nextTurn=turnList.get(aktTurnIndex);
			lastTurn=null;
			if(turnList.indexOf(nextTurn)+1<turnList.size())
				next2Turn=turnList.get(turnList.indexOf(nextTurn)+1);
			else
				next2Turn=null;
			isTurnFront=true;
		}
	}
	
	/**
	 * iterovanie zakrutami
	 */
	public void iterate(SensorModel sm) {
		if(sm.getLastLapTime() != 0 && !isInitTurnList()) { //inicializacia zoznamu zakrut ak neexistuje ulozeny subor
			initTurnList();
		}
		
		if(sm.getDistanceFromStartLine()<2 && (isLoadData || sm.getLastLapTime()!=0)) {//obnovenie iteratora zoznamu zakrut po prejdeni cielom
			lastTurnEnd=0;
			aktTurnIndex=0;
			nextTurn=turnList.get(aktTurnIndex);
			lastTurn=null;
			if(turnList.indexOf(nextTurn)+1<turnList.size())
				next2Turn=turnList.get(turnList.indexOf(nextTurn)+1);
			else
				next2Turn=null;
			isTurnFront=true;
			//System.out.println("obnovenie iteratora zoznamu zakrut");
		}
		
		//iterovanie na dalsiu zakrutu
		if((sm.getLastLapTime() != 0 || isLoadData) && aktTurnIndex+1<turnList.size() && sm.getDistanceFromStartLine()>nextTurn.getEndDistance()) {
			lastTurn=nextTurn;
			lastTurnEnd=nextTurn.getEndDistance();
			aktTurnIndex++;
			nextTurn=turnList.get(aktTurnIndex);
			if(turnList.indexOf(nextTurn)+1<turnList.size())
				next2Turn=turnList.get(turnList.indexOf(nextTurn)+1);
			else
				next2Turn=null;
			isTurnFront=true;
			/*System.out.println("peak>"+nextTurn.getPeakValue());
			System.out.println("dir>"+nextTurn.getDirection());
			System.out.println("pos>"+nextTurn.getEnterPos());
			System.out.println();*/
		}
		
		//iterovanie na dalsiu zakrutu ak aktualna je velmi mala
		if((sm.getLastLapTime() != 0 || isLoadData) && aktTurnIndex+1<turnList.size()) {
			if(nextTurn.getPeakValue()<(trackWidth>13 ? 0.08:0.05)) {
				aktTurnIndex++;
				nextTurn=turnList.get(aktTurnIndex);
				if(turnList.indexOf(nextTurn)+1<turnList.size())
					next2Turn=turnList.get(turnList.indexOf(nextTurn)+1);
				else
					next2Turn=null;
				isTurnFront=true;
				System.out.println(nextTurn.getPeakValue());
			}
		}
		
		if((sm.getLastLapTime() != 0 || isLoadData) && aktTurnIndex+1==turnList.size()) 
			isTurnFront=false;
		
		if ((eventList != null) 
				&& (eventList.size() > indexEvent+1) 
					&& ((eventList.get(indexEvent+1).getDistance() - sm.getDistanceFromStartLine()) < deviation && (eventList.get(indexEvent+1).getDistance() - sm.getDistanceFromStartLine()) > 0)){
						actEvent=eventList.get(indexEvent+1);
						//System.out.println(actEvent.getBrake()+"<----------------------------");
						indexEvent++;
		} else if ((jumpList != null) 
				&& (jumpList.size() > indexJump+1) 
					&& ((jumpList.get(indexJump+1).getDistance() - sm.getDistanceFromStartLine()) < deviation && (jumpList.get(indexJump+1).getDistance() - sm.getDistanceFromStartLine()) > 0)){
						actEvent=jumpList.get(indexJump+1);
						indexJump++;
		} else if ((actEvent != null) && ((actEvent.getDistance() - sm.getDistanceFromStartLine()) >= deviation && (actEvent.getDistance() - sm.getDistanceFromStartLine()) <= 0))
			actEvent = null;

	}

	public EventInfo getActEvent() {
		return actEvent;
	}
	
	
	
}
