package controller.model;

import java.io.Serializable;

import controller.DriveModul;

/**
 * immutable
 * @author Ivan Valencik
 */
public class Turn implements Serializable
{
	private double peakVal, startDist, endDist, speed, frontEdgeDist, enterPos, enterKm, breakDist;
	private Direction dir;
	

	public enum Direction { STRAIGHT, RIGHT, LEFT };

	public Turn(double startDistance, double endDistance, double peakValue,
			double peakSpeed, double frontEdgeDistance)
	{
		peakVal = Math.abs(peakValue);
		if (peakValue < 0)
			dir = Direction.RIGHT;
		else
			dir = Direction.LEFT;
		startDist = startDistance;
		endDist = endDistance;
		speed = peakSpeed;
		frontEdgeDist = frontEdgeDistance;
	}
	
	/**
	 * metoda, ktora nasetuje vsetkym zakrutam vstupne pozicie
	 * @param trackWidth
	 */
	public void setEnterPosition(double trackWidth, double friction) {
		if(trackWidth>14) {
			if(this.getPeakValue()<0.3) {
				this.enterPos=((trackWidth-4)/trackWidth)*(this.getPeakValue()+0.7);
				this.enterKm=this.getStartDistance()-3*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}
			else  /*if(this.getPeakValue()<0.6)*/ {
				this.enterPos=((trackWidth-4)/trackWidth)*((this.getPeakValue()+0.65)>1 ? 1:(this.getPeakValue()+0.65));
				this.enterKm=this.getStartDistance()-1.8*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}
			/*else {
				this.enterPos=(trackWidth-6)/trackWidth;
				this.enterKm=this.getStartDistance()-2*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}*/
		}
		else {
			if(this.getPeakValue()<0.2) {
				this.enterPos=((trackWidth-3)/trackWidth)*(this.getPeakValue()+0.8);
				this.enterKm=this.getStartDistance()-3*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}
			else  /*if(this.getPeakValue()<0.6)*/ {
				this.enterPos=((trackWidth-3)/trackWidth)*((this.getPeakValue()+0.65)>1 ? 1:(this.getPeakValue()+0.65));
				this.enterKm=this.getStartDistance()-3*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}
			/*else {
				this.enterPos=(trackWidth-5)/trackWidth;
				this.enterKm=this.getStartDistance()-2*trackWidth;
				if(this.getDirection()==Direction.LEFT)
					this.enterPos=-this.enterPos;
			}*/
		}
		
		if(getPeakValue()>(trackWidth>14 ? 0.15:0.1))
			setBreakDist(DriveModul.countDistanceFromSpeed(getSpeed(), friction));
		else
			setBreakDist(200);
	}

	public double getSpeed() {
		return speed;
	}

	public void setSpeed(double speed) {
		this.speed = speed;
	}

	public double getBreakDist() {
		return breakDist;
	}

	public void setBreakDist(double breakDist) {
		this.breakDist = breakDist;
	}

	public double getEnterPos() {
		return enterPos;
	}

	public void setEnterPos(double enterPos) {
		this.enterPos = enterPos;
	}

	public double getEnterKm() {
		return enterKm;
	}

	public void setEnterKm(double enterKm) {
		this.enterKm = enterKm;
	}

	public double getPeakValue() {
		return peakVal;
	}
	
	public double getPeakSpeed() {
		return speed;
	}

	public double getStartDistance() {
		return startDist;
	}

	public Direction getDirection() {
		return dir;
	}
	
	public double getEndDistance()
	{
		return endDist;
	}
	
	public double getFrontEdgeDistance() {
		return frontEdgeDist;
	}

	public String toString(){
		return dir.toString()+" : "+getPeakValue() +'\n'+"zaciatok: "+getStartDistance()+'\t'+'\t'+"konec: "+getEndDistance()+'\n';
	}
}
